package raytracing;

import basics.math.algebra.Ray;
import basics.math.algebra.Vector;
import geometry.scene.Hit;
import geometry.scene.Scene;
import optics.light.Color;
import optics.light.LightSource;
import optics.view.Camera;
import optics.view.PinholeCamera;
import renderer.Renderer;

import java.util.Optional;

public class Raytracer implements Renderer {

    private Camera camera = new PinholeCamera(new Vector(0,0,-4), Vector.origin(), 90, 1e-3, 1e3);
    private Scene scene = Scene.generateExampleScene();

    private double traceShadow(Vector point, Vector direction, double distance) {
        Ray shadowRay = new Ray(point, direction, 1e-3, distance);

        Optional<Hit> result = scene.intersect(shadowRay);

        if (result.isEmpty()) {
            return 1.0;
        }

        return 0.0;
    }

    private Color traceDirect(Ray directRay, int depth) {
        if (depth == 4) {
            return Color.BLACK;
        }

        Optional<Hit> result = scene.intersect(directRay);

        if (result.isEmpty()) {
            // Background color
            return Color.BLACK;
        }

        Vector intersection = directRay.travel(result.get().getDistance());
        Vector normal = result.get().getMesh().normalAt(intersection);

        BasicMaterial material = (BasicMaterial) result.get().getMesh().getMaterial();

        // indirect light sum
        Color incoming = new Color(0,0,0);

        Color reflectedLight = reflectedLight(directRay, material, normal, intersection, depth);

        for (LightSource light : scene.getLights()) {
            double directInfluence = castShadow(light, intersection, normal);

            Color directLight = material.getDiffuse().scale(directInfluence).mul(light.getColor());

            incoming = incoming.add(directLight).add(reflectedLight);
        }

        return incoming.add(reflectedLight);
    }

    private Color reflectedLight(Ray incomingRay, BasicMaterial material, Vector normal, Vector point, int depth) {
        Vector reflected = incomingRay.getDirection().reflect(normal);

        Color reflectedLight = traceDirect(new Ray(point, reflected, 1e-3, 1e3), depth + 1);

        double highlightFactor = reflected.dot(normal);

        return reflectedLight.add(new Color(highlightFactor)).mul(material.getReflectionTint());
    }

    private double castShadow(LightSource target, Vector point, Vector surfaceNormal) {
        Vector lightDirection = target.getDirection(point);
        double distance = target.getDistance(point);
        double shadow = traceShadow(point.add(surfaceNormal.scale(1e-3)), lightDirection, distance);
        return Math.max(surfaceNormal.dot(lightDirection), 0.0) * shadow;
    }

    @Override
    public Color traceCameraRay(double u, double v) {
        Ray viewRay = camera.generateViewRay(u, v);

        return traceDirect(viewRay, 0);
    }

    @Override
    public Camera getCamera() {
        return camera;
    }
}
